from __future__ import annotations

class Pipeline:
    '''Engine Rendering Pipeline'''

    PYGAME:int = 0
    '''
    PyGame will render all elements/objects
    '''

    OPENGL:int = 1
    '''
    OpenGL will render all elements/objects
    '''