'''
## Wame Engine
Simple, Pythonic, Intuitive, Pygame Wrapper

#### Version: `0.4.0`
#### Documentation: https://github.com/WilDev-Studios/Wame
'''

from wame.engine import Engine
from wame.pipeline import Pipeline
from wame.scene import Scene
from wame.settings import Settings

from wame import color
from wame import common
from wame import ui
from wame import vector