from __future__ import annotations

from wame.color.rgb import ColorRGB

class RGB:
    '''Common RGB Values'''

    BLACK:ColorRGB = ColorRGB(0, 0, 0)
    '''`RGB: 0, 0, 0`'''

    BLUE:ColorRGB = ColorRGB(0, 0, 255)
    '''`RGB: 0, 0, 255`'''

    CYAN:ColorRGB = ColorRGB(0, 255, 255)
    '''`RGB: 0, 255, 255`'''

    GREEN:ColorRGB = ColorRGB(0, 255, 0)
    '''`RGB: 0, 255, 0`'''

    PINK:ColorRGB = ColorRGB(255, 0, 255)
    '''`RGB: 255, 0, 255`'''

    RED:ColorRGB = ColorRGB(255, 0, 0)
    '''`RGB: 255, 0, 0`'''

    WHITE:ColorRGB = ColorRGB(255, 255, 255)
    '''`RGB: 255, 255, 255`'''

    YELLOW:ColorRGB = ColorRGB(255, 255, 0)
    '''`RGB: 255, 255, 0`'''