'''
## Vector Module - Wame Engine
Contains all necessary vector objects for 2D, 3D, and 4D positional, sizeable, rotational, and color fields
'''

from wame.vector.xy import FloatVector2, IntVector2
from wame.vector.xyz import FloatVector3, IntVector3