'''
## Common Module - Wame Engine
Contains all objects and functionality that might be useful to developers
'''

from wame.common.keys import KEYS, is_char, is_letter, is_lower, is_number, is_symbol, is_upper